/*
  ==============================================================================

    SpectroscopeViewer.cpp
    Created: 27 Jul 2014 2:34:48pm
    Author:  Samuel Gaehwiler

    Heavily based on the dRowAudio Spectroscope class by David Rowland.
 
  ==============================================================================
*/

#include "SpectrumViewer.h"

#if JUCE_MAC || JUCE_IOS || DROWAUDIO_USE_FFTREAL


// static member initialisation
// ============================

const int SpectrumViewer::frequenciesToPlot[] = {10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 2000, 3000, 4000, 5000, 6000, 7000, 8000, 9000, 10000, 20000};
const int SpectrumViewer::numberOfFrequenciesToPlot = 29;


//==============================================================================
SpectrumViewer::SpectrumViewer (Value& repaintViewerValue,
                                drow::Buffer& magnitudeBuffer,
                                Value& detectedFrequencyValue)
  : sampleRate                {44100.0},
    repaintViewer             (var(false)),
    fftMagnitudeBuffer        (magnitudeBuffer),
    detectedFrequency         {var(0)},
    frequencyToDisplay        {var(0)},
    mouseMode                 {false},
    mouseXPosition            {0},
    heightForFrequencyCaption {20},
    gradientImage             (Image::RGB, 100, 100, false)
{
	setOpaque (true);
    
    repaintViewer.referTo (repaintViewerValue);
    
    detectedFrequency.referTo (detectedFrequencyValue);
    detectedFrequency.addListener(this);

    gradientImage.clear (gradientImage.getBounds(), Colours::black);

    addAndMakeVisible (&frequencyCaption);
    
    path571 = get571RecordingStudioPath();
    
    startTimer (30);
}


SpectrumViewer::~SpectrumViewer()
{
}


void SpectrumViewer::resized()
{
    gradientImage = Image (Image::RGB, jmax (1, getWidth()), jmax (1, getHeight() - heightForFrequencyCaption), false);
    createGradientImage();

    frequencyCaption.setBounds (0, getHeight() - heightForFrequencyCaption,
                                 getWidth(), heightForFrequencyCaption);
}


void SpectrumViewer::paint(Graphics& g)
{
    const int w = getWidth();
    const int h = getHeight() - heightForFrequencyCaption;
    
    g.setColour (Colours::black);
    g.fillRect(0, 0, w, h);
    
    
    // Draw the "571 recording studio" logo
    // ------------------------------------
    Rectangle<float> targetRectangle = getLocalBounds().toFloat();
    targetRectangle.removeFromBottom (heightForFrequencyCaption);
    targetRectangle = targetRectangle.reduced(10.0f);
    // g.setColour (Colours::black.brighter(0.125));
    g.setColour (Colours::lightgoldenrodyellow.darker(JUCE_LIVE_CONSTANT (7.5f)));
    g.fillPath (path571, RectanglePlacement (RectanglePlacement::centred)
                .getTransformToFit (path571.getBounds(), targetRectangle));
    //    g.setColour (Colours::green);
    //    float lineThickness2 = 0.75f;
    //    g.strokePath (path571, PathStrokeType(lineThickness2), RectanglePlacement (RectanglePlacement::centred)
    //                  .getTransformToFit (path571.getBounds(), targetRectangle));
    
    
    // Draw the vertical lines
    // -----------------------
    g.setColour (Colours::darkgreen.darker().darker());
    
    for (int i = 0; i < numberOfFrequenciesToPlot; ++i)
    {
        const double proportion = frequenciesToPlot[i] / (sampleRate * 0.5);
        int xPos = logTransformInRange0to1 (proportion) * getWidth();
        g.drawVerticalLine(xPos, 0.0f, getHeight());
    }
    
    
    // Draw the actual result from the FFT
    // -----------------------------------
    const int numBins = fftMagnitudeBuffer.getSize();
    const float* data = fftMagnitudeBuffer.getData();
    
    // NOTE TO DAVE96: The jlimit is not needed. Puts the load off the CPU by 1-2%.
    // Original line:      float y2, y1 = jlimit (0.0f, 1.0f, float (1 + (drow::toDecibels (data[0]) / 100.0f)));
    float x = 0;
    float y = 0;
    
    // The path which will be the border of the filled area.
    Path spectrumPath;
    // Add the top left point.
    // Only values > 0 are passed to drow::toDecibels().
    const float yInPercent = data[0]>0 ? float (1 + (drow::toDecibels (data[0]) / 100.0f)) : -0.01;
    y = h - h * yInPercent;
    // No coordinate should be NaN
    jassert (x == x && y == y);
    spectrumPath.startNewSubPath(x, y);
    
    for (int i = 1; i < numBins; ++i)
    {
        x = logTransformInRange0to1 ((float)i / numBins) * w;
        // Only values > 0 are passed to drow::toDecibels().
        const float yInPercent = data[i]>0 ? float (1 + (drow::toDecibels (data[i]) / 100.0f)) : -0.01;
        y = h - h * yInPercent;
        
        spectrumPath.lineTo(x, y);
    }
    
    g.setColour (Colours::green);
    float lineThickness = 0.75f;
    g.strokePath (spectrumPath, PathStrokeType(lineThickness));
    
    // Remark: Using g.drawVerticalLine - which is told to be much more efficient for the renderer -
    // on every bin position instead of using a Path improved the overall performance of this
    // spectrum analyser by only 5%. (I profiled it)
    
    // Bottom right point.
    spectrumPath.lineTo(x, getHeight());
    // Bottom left point.
    spectrumPath.lineTo(0, getHeight());
    spectrumPath.closeSubPath();
    
    // Fill the graph.
    const float opacity = 0.8f;
    g.setTiledImageFill(gradientImage, 0, 0, opacity);
    g.fillPath(spectrumPath);
    
    
    // Draw a vertical line at the mouse position
    // ------------------------------------------
    if (mouseMode)
    {
        g.setColour (Colours::lightgoldenrodyellow);
        g.drawVerticalLine(mouseXPosition -1, 0, getHeight());
        // Don't know why the -1 offset for the x position
        // is needed. (It is also needed if the drawing
        // would be placed directly on the component, i.e.
        // in the paint() function.)
        // Maybe because of rounding in the mouseMove function.
    }
    else
    {
        const double proportion = (double)detectedFrequency.getValue() / (sampleRate / 2.0);
        const int pitchXCoord = roundToInt (logTransformInRange0to1 (proportion) * getWidth());
        
        g.setColour (Colours::green);
        g.drawVerticalLine (pitchXCoord, 0.0f, (float) getHeight());
    }

}


//==============================================================================
void SpectrumViewer::setSampleRate (double newSampleRate)
{
    sampleRate = newSampleRate;
    frequencyCaption.setSampleRate (newSampleRate);
}


int SpectrumViewer::getHeightOfFrequencyCaption()
{
    return heightForFrequencyCaption;
}


Value & SpectrumViewer::getFrequencyToDisplay()
{
    return frequencyToDisplay;
}


void SpectrumViewer::timerCallback()
{
    // TODO: How compares the CPU usage if we listen to repaintViewer
    //   instead of doing it here in the timerCallback?
    if (repaintViewer == true)
	{
        repaint(0, 0, getWidth(), getHeight() - heightForFrequencyCaption);
            // - heightForFrequencyCaption: We don't need to repaint the frequency caption.
        
        repaintViewer = false;
    }

	// Make the FFT levels fall.
    const int magnitudeBufferSize = fftMagnitudeBuffer.getSize();
	float* magnitudeBuffer = fftMagnitudeBuffer.getData();
    
	for (int i = 0; i < magnitudeBufferSize; i++)
		magnitudeBuffer[i] *= JUCE_LIVE_CONSTANT (0.707f);
}


void SpectrumViewer::valueChanged (Value &value)
{
    if (!mouseMode && value.refersToSameSourceAs(detectedFrequency))
    {
        frequencyToDisplay = detectedFrequency.getValue();
    }
}


void SpectrumViewer::mouseEnter (const MouseEvent &event)
{
    mouseMode = true;
}


void SpectrumViewer::mouseMove (const MouseEvent &event)
{
    mouseXPosition = event.getPosition().getX();
    
    const float normalizedXPosition = (mouseXPosition) / (float)getWidth();
    const float frequency = sampleRate / 2.0f * expTransformInRange0to1 (normalizedXPosition);
    frequencyToDisplay = (int)frequency;
}


void SpectrumViewer::mouseExit (const MouseEvent &event)
{
    mouseMode = false;
}

void SpectrumViewer::createGradientImage()
{
    Graphics g (gradientImage);
    ColourGradient gradient (Colours::darkgreen.darker(), 0.0f, 0.8f * gradientImage.getHeight(), Colours::black, 0.0f, gradientImage.getHeight(), false);
    g.setGradientFill(gradient);
    g.fillAll();
}

Path SpectrumViewer::get571RecordingStudioPath()
{
    static const unsigned char pathData[] = { 110,109,2,171,30,66,217,206,74,66,98,250,254,33,66,229,80,75,66,188,116,39,66,199,203,76,66,27,175,43,66,92,15,78,66,98,52,179,56,66,207,247,81,66,68,139,70,66,114,104,88,66,160,26,81,66,127,106,95,66,98,160,26,86,66,100,187,98,66,156,68,91,66,248,211,
        101,66,193,74,96,66,150,67,106,66,98,56,9,101,66,183,115,110,66,193,74,105,66,254,212,114,66,218,78,109,66,175,71,119,66,98,60,95,113,66,169,198,123,66,44,7,117,66,86,78,128,66,52,179,120,66,164,176,130,66,98,136,214,131,66,121,105,140,66,199,203,136,
        66,107,124,154,66,148,24,137,66,113,189,169,66,98,109,103,137,66,41,92,185,66,156,4,133,66,37,134,199,66,122,105,123,66,96,37,210,66,98,24,89,122,66,133,235,210,66,249,83,121,66,223,143,211,66,241,39,120,66,229,80,212,66,98,155,25,119,66,131,0,213,66,
        208,119,118,66,86,206,213,66,171,113,117,66,106,124,214,66,98,73,97,115,66,172,220,215,66,214,120,114,66,149,3,216,66,3,43,112,66,129,85,217,66,98,193,202,111,66,199,139,217,66,95,58,110,66,96,229,218,66,30,218,109,66,147,24,219,66,98,120,62,107,66,131,
        128,220,66,67,224,102,66,20,46,222,66,75,140,99,66,188,180,223,66,98,128,234,98,66,125,255,223,66,100,144,98,66,137,129,224,66,141,236,97,66,61,202,224,66,98,243,210,92,66,210,13,227,66,120,190,85,66,225,122,229,66,239,124,79,66,235,145,231,66,98,178,
        242,55,66,20,110,239,66,222,36,21,66,128,21,244,66,83,141,217,65,128,85,241,66,98,89,14,201,65,161,197,240,66,124,233,184,65,136,129,239,66,110,188,168,65,122,20,238,66,98,99,143,120,65,120,41,234,66,204,32,46,65,153,217,227,66,192,157,223,64,52,94,219,
        66,98,159,237,212,64,214,163,218,66,233,249,202,64,208,226,217,66,139,63,193,64,50,51,217,66,98,34,174,171,64,132,171,215,66,110,184,118,64,187,116,212,66,147,190,119,64,226,37,210,66,98,243,163,120,64,210,13,208,66,237,79,169,64,244,104,205,66,160,237,
        188,64,150,46,204,66,98,178,112,213,64,101,166,202,66,213,75,235,64,207,162,201,66,24,88,3,65,15,216,199,66,98,75,139,14,65,195,96,198,66,237,208,30,65,203,140,196,66,110,102,42,65,38,177,194,66,98,235,165,45,65,7,44,194,66,235,165,49,65,188,116,193,
        66,98,100,55,65,67,203,192,66,98,245,124,63,65,178,221,191,66,10,43,77,65,59,223,189,66,180,28,90,65,92,207,189,66,98,88,141,111,65,194,181,189,66,55,221,122,65,84,227,192,66,14,215,131,65,39,177,194,66,98,164,26,138,65,189,116,196,66,170,155,146,65,
        14,45,198,66,71,139,153,65,16,216,199,66,98,44,92,165,65,39,177,202,66,106,102,176,65,241,82,204,66,145,151,192,65,28,90,206,66,98,237,38,201,65,14,109,207,66,219,163,209,65,237,188,208,66,194,159,218,65,3,86,209,66,98,210,247,229,65,134,22,210,66,6,
        43,242,65,249,254,209,66,20,88,253,65,226,37,210,66,98,217,35,25,66,58,223,210,66,233,251,46,66,130,192,206,66,184,243,59,66,109,146,199,66,98,168,155,63,66,203,140,197,66,65,53,68,66,246,19,195,66,213,205,70,66,191,202,192,66,98,59,52,78,66,173,71,186,
        66,238,81,84,66,144,109,174,66,168,27,81,66,134,150,164,66,98,57,9,79,66,249,62,158,66,225,79,73,66,156,47,152,66,209,119,66,66,228,80,148,66,98,197,245,65,66,48,8,148,66,55,94,65,66,78,205,147,66,6,214,64,66,130,128,147,66,98,125,20,63,66,124,127,146,
        66,164,69,61,66,173,7,145,66,67,224,58,66,218,249,143,66,98,59,52,55,66,183,94,142,66,212,205,49,66,73,140,140,66,225,79,45,66,15,24,139,66,98,30,90,37,66,23,132,136,66,172,113,25,66,52,30,134,66,225,79,13,66,195,32,133,66,98,10,172,2,66,142,66,132,66,
        77,12,239,65,160,197,131,66,54,51,213,65,160,197,131,66,98,177,71,177,65,160,197,131,66,60,180,140,65,160,197,131,66,190,30,77,65,160,197,131,66,98,212,247,55,65,160,197,131,66,32,47,19,65,185,9,132,66,65,223,9,65,62,245,130,66,98,24,2,255,64,142,194,
        129,66,69,53,2,65,192,245,125,66,61,137,5,65,159,69,119,66,98,241,81,22,65,249,41,85,66,100,186,37,65,216,249,42,66,166,26,51,65,6,44,9,66,98,156,67,55,65,108,61,253,65,182,114,60,65,131,22,233,65,193,73,66,65,14,131,212,65,98,170,112,77,65,22,47,173,
        65,144,65,84,65,26,133,129,65,242,81,92,65,36,178,49,65,98,56,8,94,65,202,77,32,65,35,90,98,65,93,102,10,65,226,249,100,65,66,227,233,64,98,253,40,104,65,17,219,181,64,105,16,104,65,122,108,139,64,7,0,114,65,234,44,114,64,98,112,231,130,65,0,6,41,64,
        28,4,177,65,160,32,80,64,6,43,187,65,201,124,79,64,98,16,2,251,65,250,132,75,64,102,59,40,66,252,175,66,64,20,131,73,66,201,124,79,64,98,202,118,83,66,211,83,83,64,69,11,103,66,213,126,50,64,210,34,107,66,234,44,114,64,98,169,198,109,66,86,145,141,64,
        247,40,109,66,156,71,189,64,218,78,109,66,94,61,238,64,98,99,144,109,66,187,32,34,65,206,204,109,66,20,90,62,65,218,78,109,66,171,200,114,65,98,124,20,109,66,137,151,133,65,52,179,109,66,245,126,144,65,210,34,107,66,137,151,149,65,98,149,152,103,66,211,
        163,156,65,140,236,88,66,141,237,153,65,226,122,79,66,141,237,153,65,98,64,53,54,66,141,237,153,65,152,238,30,66,141,237,153,65,76,183,5,66,141,237,153,65,98,80,13,0,66,141,237,153,65,21,174,245,65,112,147,152,65,211,77,240,65,143,24,156,65,98,195,245,
        231,65,100,145,161,65,176,114,231,65,145,67,197,65,201,118,229,65,137,151,213,65,98,11,215,222,65,170,28,6,66,37,6,219,65,233,81,27,66,174,71,214,65,160,69,55,66,98,141,151,213,65,70,97,59,66,190,159,209,65,33,91,67,66,51,51,213,65,187,116,70,66,98,166,
        155,218,65,207,34,75,66,0,0,244,65,126,106,72,66,72,97,1,66,189,159,72,66,109,29,218,14,67,227,165,175,65,98,148,88,15,67,223,79,188,65,142,23,16,67,198,75,220,65,2,75,17,67,67,139,225,65,98,250,30,18,67,1,43,229,65,129,117,19,67,112,61,227,65,104,209,
        20,67,67,139,225,65,98,204,236,22,67,108,231,222,65,206,55,24,67,63,53,219,65,241,178,25,67,167,198,215,65,98,243,253,26,67,179,200,212,65,137,225,29,67,83,227,202,65,235,113,30,67,24,4,206,65,98,208,34,31,67,22,217,209,65,219,217,30,67,108,231,236,65,
        219,217,30,67,114,104,249,65,98,219,217,30,67,61,138,68,66,76,183,30,67,217,142,132,66,76,183,30,67,110,146,167,66,98,76,183,30,67,32,112,184,66,76,183,30,67,192,138,197,66,76,183,30,67,237,124,214,66,98,76,183,30,67,200,246,218,66,84,35,30,67,172,156,
        234,66,78,66,31,67,249,254,236,66,98,232,59,32,67,241,18,239,66,191,127,36,67,28,90,238,66,219,217,38,67,28,90,238,66,98,221,4,41,67,28,90,238,66,250,94,45,67,141,87,239,66,235,113,46,67,249,254,236,66,98,221,36,47,67,206,119,235,66,237,252,46,67,245,
        104,228,66,237,252,46,67,63,181,223,66,98,237,252,46,67,82,56,189,66,237,252,46,67,118,126,155,66,237,252,46,67,171,28,117,66,98,237,252,46,67,77,226,48,66,237,252,46,67,134,22,221,65,237,252,46,67,96,59,43,65,98,237,252,46,67,67,225,4,65,98,48,47,67,
        186,245,152,64,235,113,46,67,250,1,123,64,98,235,113,45,67,156,71,49,64,214,35,38,67,121,108,79,64,104,209,36,67,160,157,79,64,98,94,122,35,67,199,206,79,64,8,12,33,67,78,229,64,64,204,204,31,67,119,65,88,64,98,108,71,30,67,29,221,116,64,231,91,29,67,
        182,159,178,64,102,70,28,67,148,239,207,64,98,100,27,28,67,216,122,212,64,245,8,28,67,19,90,216,64,118,222,27,67,185,245,220,64,98,224,186,24,67,59,53,26,65,112,189,20,67,24,90,66,65,253,52,16,67,176,200,88,65,98,124,159,15,67,16,174,91,65,28,250,14,
        67,145,67,93,65,235,113,14,67,184,116,97,65,98,116,19,13,67,231,81,108,65,6,1,13,67,233,124,117,65,106,92,13,67,16,131,135,65,98,2,203,13,67,195,245,150,65,190,95,14,67,66,96,163,65,29,218,14,67,227,165,175,65,99,109,100,59,145,66,29,90,238,66,98,157,
        239,151,66,187,137,238,66,55,137,156,66,29,90,238,66,135,150,162,66,29,90,238,66,98,80,77,165,66,29,90,238,66,123,212,168,66,246,168,238,66,82,184,170,66,92,207,237,66,98,84,99,173,66,172,156,236,66,43,135,175,66,205,140,230,66,31,197,177,66,227,37,226,
        66,98,248,83,184,66,49,72,213,66,15,109,190,66,76,55,200,66,232,123,196,66,57,116,188,66,98,175,135,197,66,2,107,186,66,183,115,198,66,223,79,184,66,201,118,199,66,75,55,182,66,98,232,123,200,66,40,28,180,66,8,172,201,66,98,16,178,66,207,183,202,66,93,
        250,175,66,98,191,223,204,66,169,177,171,66,209,162,206,66,238,103,167,66,158,175,208,66,99,59,163,66,98,150,195,210,66,5,1,159,66,207,247,212,66,167,6,155,66,140,236,214,66,11,194,150,66,98,203,225,216,66,230,123,146,66,99,80,219,66,226,101,142,66,152,
        110,221,66,179,72,138,66,98,185,158,223,66,173,7,134,66,162,69,225,66,48,136,129,66,97,101,227,66,114,19,123,66,98,122,169,229,66,83,142,114,66,115,168,231,66,61,53,106,66,79,162,233,66,194,32,98,66,98,112,210,239,66,245,211,72,66,20,67,246,66,12,173,
        47,66,18,88,252,66,104,188,22,66,98,24,89,253,66,194,160,18,66,45,71,254,66,94,101,14,66,249,83,255,66,141,66,10,66,98,154,57,0,67,8,215,5,66,138,225,0,67,42,221,255,65,15,109,1,67,231,81,248,65,98,42,252,1,67,149,153,240,65,244,93,2,67,210,163,233,65,
        115,200,2,67,186,159,226,65,98,82,88,3,67,106,18,217,65,221,4,4,67,41,178,205,65,133,139,4,67,135,108,198,65,98,104,209,4,67,235,167,194,65,6,225,4,67,106,18,192,65,135,22,5,67,248,169,188,65,98,141,215,5,67,170,71,176,65,10,23,7,67,76,141,161,65,155,
        100,7,67,174,157,149,65,98,139,204,7,67,137,151,133,65,42,135,7,67,198,247,93,65,42,135,7,67,96,145,47,65,98,42,135,7,67,200,34,7,65,15,248,7,67,198,163,152,64,171,252,6,67,178,77,114,64,98,62,53,6,67,242,130,64,64,148,227,1,67,32,96,77,64,187,52,0,67,
        145,157,79,64,98,114,83,240,66,178,77,90,64,69,33,224,66,145,157,79,64,92,58,209,66,145,157,79,64,98,156,111,188,66,145,157,79,64,61,181,164,66,247,3,62,64,90,79,146,66,145,157,79,64,98,82,163,143,66,237,44,82,64,65,11,141,66,145,157,79,64,78,205,139,
        66,254,44,134,64,98,49,115,138,66,35,51,167,64,141,66,139,66,54,10,25,65,141,66,139,66,90,16,54,65,98,141,66,139,66,80,57,88,65,202,76,138,66,31,219,139,65,145,88,140,66,39,135,148,65,98,182,222,141,66,227,251,154,65,106,103,146,66,178,243,153,65,100,
        166,150,66,178,243,153,65,98,73,183,175,66,178,243,153,65,133,86,197,66,178,243,153,65,86,249,221,66,178,243,153,65,98,63,160,225,66,178,243,153,65,106,39,232,66,76,141,152,65,139,23,233,66,59,53,157,65,98,194,160,233,66,55,223,159,65,82,35,232,66,100,
        145,173,65,168,49,231,66,106,18,181,65,98,156,239,227,66,239,253,206,65,137,172,224,66,32,6,232,65,217,249,221,66,250,212,254,65,98,24,175,217,66,35,134,17,66,223,122,212,66,178,200,37,66,250,105,208,66,43,50,54,66,98,21,89,207,66,16,131,58,66,33,91,
        206,66,113,232,62,66,25,111,205,66,6,172,66,66,98,64,96,201,66,135,65,83,66,27,26,197,66,0,43,100,66,224,58,193,66,41,7,116,66,98,206,183,191,66,55,52,122,66,25,239,189,66,91,79,129,66,21,89,188,66,73,12,132,66,98,236,60,187,66,75,247,133,66,68,118,186,
        66,112,189,135,66,82,163,185,66,212,120,137,66,98,117,190,182,66,73,140,143,66,76,162,179,66,79,13,150,66,31,176,176,66,251,233,155,66,98,242,125,172,66,5,65,164,66,252,84,168,66,212,120,172,66,230,123,164,66,140,151,180,66,98,250,169,163,66,103,81,182,
        66,201,97,162,66,154,68,184,66,29,197,161,66,117,190,185,66,98,2,150,161,66,37,49,186,66,64,96,161,66,41,71,186,66,224,58,161,66,215,142,186,66,98,230,251,158,66,51,222,190,66,32,48,157,66,47,200,195,66,242,253,154,66,240,146,199,66,98,246,211,154,66,
        33,219,199,66,195,224,154,66,215,14,200,66,205,183,154,66,82,99,200,66,98,207,226,153,66,164,27,202,66,164,27,153,66,211,120,203,66,172,71,152,66,123,255,204,66,98,225,229,148,66,86,57,211,66,125,42,146,66,125,106,218,66,184,201,142,66,129,64,224,66,
        98,202,12,142,66,35,134,225,66,194,96,141,66,35,6,227,66,213,227,140,66,72,12,228,66,98,176,157,140,66,176,157,228,66,182,158,140,66,58,95,229,66,20,89,140,66,43,242,229,66,98,233,209,139,66,197,11,231,66,59,10,139,66,186,244,231,66,79,184,138,66,18,
        238,232,66,98,111,18,137,66,82,248,237,66,80,205,139,66,182,51,238,66,100,59,145,66,29,90,238,66,99,109,184,110,132,67,86,142,133,66,98,241,114,129,67,86,142,133,66,196,0,125,67,20,238,132,66,87,25,122,67,164,240,133,66,98,28,250,121,67,101,251,133,66,
        200,214,121,67,14,237,133,66,9,183,121,67,164,240,133,66,98,238,71,120,67,10,23,134,66,21,217,118,67,160,26,136,66,21,57,119,67,248,19,140,66,98,56,148,119,67,178,221,143,66,121,244,121,67,254,212,143,66,23,4,124,67,252,233,143,66,98,48,40,126,67,125,
        255,143,66,91,159,128,67,39,177,142,66,220,68,129,67,21,174,144,66,98,109,2,130,67,201,246,146,66,200,166,129,67,132,128,161,66,200,166,129,67,46,178,167,66,98,200,166,129,67,9,44,186,66,200,166,129,67,240,231,198,66,200,166,129,67,224,143,217,66,98,
        200,166,129,67,97,101,223,66,40,92,129,67,171,241,231,66,98,240,129,67,242,210,234,66,98,47,45,130,67,0,0,236,66,243,157,130,67,181,200,236,66,75,23,131,67,244,189,236,66,98,5,209,132,67,4,150,236,66,151,110,132,67,64,245,226,66,151,110,132,67,224,143,
        217,66,98,151,110,132,67,232,123,199,66,151,110,132,67,136,150,185,66,151,110,132,67,124,20,168,66,98,151,110,132,67,89,185,162,66,139,28,132,67,212,205,146,66,6,209,132,67,22,174,144,66,98,129,101,133,67,153,238,142,66,186,249,134,67,151,3,144,66,156,
        68,136,67,253,233,143,66,98,97,133,137,67,99,208,143,66,123,180,138,67,71,182,143,66,127,170,138,67,194,138,138,66,98,139,156,138,67,217,78,131,66,201,150,134,67,86,142,133,66,184,110,132,67,86,142,133,66,99,109,137,65,134,67,78,98,148,64,98,227,117,
        132,67,78,98,148,64,55,153,130,67,76,55,145,64,215,19,129,67,78,98,148,64,98,149,67,128,67,74,12,150,64,182,147,126,67,51,51,143,64,53,190,125,67,62,10,179,64,98,67,11,125,67,136,22,209,64,2,43,125,67,226,122,16,65,2,43,125,67,130,149,53,65,98,2,43,125,
        67,176,114,181,65,2,43,125,67,98,16,1,66,2,43,125,67,180,72,46,66,98,2,43,125,67,12,130,60,66,162,133,124,67,127,234,88,66,217,238,125,67,178,157,94,66,98,188,212,126,67,100,59,98,66,100,251,128,67,26,175,97,66,98,240,129,67,26,175,97,66,98,186,41,131,
        67,26,175,97,66,200,54,132,67,26,175,97,66,123,100,133,67,26,175,97,66,98,10,87,135,67,26,175,97,66,186,9,139,67,77,98,102,66,64,37,139,67,225,122,88,66,98,230,64,139,67,231,123,74,66,246,56,137,67,219,249,76,66,162,5,135,67,219,249,76,66,98,109,151,
        133,67,219,249,76,66,232,11,130,67,156,68,79,66,119,142,129,67,8,172,74,66,98,72,17,129,67,147,24,70,66,80,93,129,67,137,65,46,66,80,93,129,67,229,80,36,66,98,80,93,129,67,118,190,30,66,80,93,129,67,206,247,22,66,80,93,129,67,53,94,16,66,98,80,93,129,
        67,33,48,10,66,211,61,129,67,49,8,0,66,119,142,129,67,65,96,250,65,98,25,84,130,67,139,108,236,65,242,2,134,67,198,75,255,65,234,54,135,67,214,163,239,65,98,250,78,135,67,126,106,238,65,111,162,135,67,54,137,231,65,171,177,135,67,95,229,228,65,98,15,
        109,136,67,40,92,196,65,136,182,133,67,160,69,202,65,110,135,132,67,228,208,202,65,98,106,65,132,67,169,241,202,65,44,247,131,67,212,120,201,65,96,170,131,67,173,71,201,65,98,45,39,131,67,193,245,200,65,28,207,129,67,24,4,204,65,164,117,129,67,220,36,
        195,65,98,53,35,129,67,169,241,186,65,49,93,129,67,144,237,147,65,49,93,129,67,107,231,139,65,98,49,93,129,67,106,231,119,65,80,18,129,67,112,104,59,65,88,142,129,67,14,88,41,65,98,49,221,129,67,32,219,29,65,248,24,131,67,194,32,32,65,217,243,131,67,
        194,32,32,65,98,158,244,133,67,194,32,32,65,215,216,138,67,165,198,49,65,246,61,139,67,16,131,4,65,98,107,81,139,67,184,116,247,64,82,45,139,67,212,206,195,64,66,37,139,67,85,100,191,64,98,117,179,138,67,55,137,129,64,90,20,136,67,78,98,148,64,137,65,
        134,67,78,98,148,64,99,109,221,228,112,67,90,164,187,66,98,225,90,112,67,127,234,186,66,35,187,111,67,8,108,186,66,192,42,111,67,88,185,185,66,98,30,165,110,67,254,20,185,66,217,46,110,67,235,17,184,66,137,161,109,67,8,108,183,66,98,4,54,108,67,0,192,
        181,66,135,214,106,67,26,47,181,66,213,56,105,67,82,248,179,66,98,84,99,104,67,148,88,179,66,55,137,103,67,121,169,178,66,29,186,102,67,80,13,178,66,98,199,43,98,67,191,159,174,66,215,35,91,67,172,92,171,66,27,175,89,67,64,181,162,66,98,105,17,89,67,
        56,9,159,66,183,115,89,67,13,194,154,66,131,224,89,67,107,188,152,66,98,113,29,91,67,17,216,146,66,219,89,95,67,7,129,143,66,141,119,99,67,97,37,143,66,98,41,60,101,67,244,253,142,66,149,35,104,67,142,215,143,66,250,94,106,67,152,174,144,66,98,60,159,
        107,67,234,38,145,66,236,145,108,67,185,222,145,66,137,161,109,67,230,16,145,66,98,215,131,110,67,222,100,144,66,209,34,111,67,91,228,142,66,192,42,111,67,70,118,140,66,98,75,55,111,67,226,58,136,66,245,232,109,67,45,178,135,66,34,123,108,67,195,181,
        134,66,98,16,56,107,67,11,215,133,66,30,133,105,67,201,182,132,66,196,224,103,67,246,104,132,66,98,214,195,101,67,25,4,132,66,36,6,99,67,105,81,132,66,190,31,98,67,246,104,132,66,98,2,11,96,67,53,158,132,66,100,27,95,67,6,65,133,66,174,231,93,67,170,
        241,133,66,98,188,180,92,67,72,161,134,66,18,131,91,67,14,237,134,66,143,66,90,67,250,62,136,66,98,28,154,88,67,113,253,137,66,182,147,86,67,80,205,141,66,200,118,85,67,105,17,145,66,98,83,35,83,67,216,227,151,66,184,190,82,67,3,107,161,66,71,129,84,
        67,76,119,168,66,98,253,52,85,67,168,70,171,66,106,252,86,67,215,163,175,66,114,136,88,67,86,14,178,66,98,155,164,89,67,61,202,179,66,36,38,91,67,31,5,181,66,157,143,92,67,37,70,182,66,98,180,136,95,67,2,235,184,66,128,117,98,67,141,23,187,66,202,97,
        101,67,21,46,189,66,98,177,61,104,67,82,56,191,66,10,119,106,67,148,24,193,66,159,122,108,67,130,21,196,66,98,249,254,109,67,5,86,198,66,116,19,111,67,134,171,200,66,73,140,111,67,163,133,204,66,98,134,86,112,67,14,2,211,66,188,52,111,67,224,79,217,66,
        18,3,110,67,1,64,220,66,98,198,235,107,67,234,102,225,66,178,29,103,67,64,53,228,66,45,178,98,67,187,137,227,66,98,92,143,93,67,13,194,226,66,211,141,90,67,101,59,222,66,213,184,88,67,150,67,215,66,98,213,88,88,67,248,211,213,66,0,32,88,67,250,190,211,
        66,176,146,87,67,121,169,210,66,98,209,66,86,67,17,24,208,66,252,169,83,67,187,9,208,66,233,198,82,67,121,169,210,66,98,149,163,81,67,49,8,214,66,114,40,83,67,211,13,220,66,40,188,83,67,2,43,222,66,98,155,100,84,67,123,148,224,66,85,142,85,67,33,240,
        226,66,106,156,86,67,242,18,229,66,98,245,200,87,67,164,112,231,66,235,113,89,67,33,48,234,66,149,163,90,67,70,54,235,66,98,131,96,96,67,72,33,240,66,237,60,106,67,199,139,239,66,81,248,110,67,224,15,234,66,98,249,62,112,67,255,148,232,66,34,251,112,
        67,83,248,230,66,251,9,114,67,165,176,228,66,98,185,201,115,67,16,237,224,66,122,52,116,67,243,82,223,66,61,234,116,67,130,85,218,66,98,170,49,117,67,60,95,216,66,141,151,117,67,183,115,214,66,217,174,117,67,124,148,212,66,98,209,34,118,67,239,60,203,
        66,158,175,116,67,48,93,196,66,125,255,114,67,54,222,191,66,98,82,56,114,67,43,199,189,66,29,250,113,67,29,26,189,66,221,228,112,67,90,164,187,66,99,109,39,81,219,67,78,98,95,66,98,219,169,219,67,70,54,97,66,70,182,220,67,25,132,97,66,193,10,221,67,234,
        38,96,66,98,37,150,221,67,109,231,93,66,54,158,221,67,31,133,85,66,54,158,221,67,3,171,74,66,98,54,158,221,67,134,107,26,66,54,158,221,67,194,202,236,65,54,158,221,67,8,129,144,65,98,54,158,221,67,10,129,111,65,68,251,221,67,94,14,233,64,156,84,221,67,
        118,18,179,64,98,80,189,220,67,22,45,130,64,160,218,218,67,212,204,144,64,101,91,218,67,134,106,148,64,98,70,102,217,67,48,92,155,64,3,91,216,67,118,18,159,64,214,136,216,67,198,159,246,64,98,183,179,216,67,24,174,35,65,132,48,218,67,145,151,20,65,64,
        165,218,67,106,102,44,65,98,11,231,218,67,38,219,57,65,187,185,218,67,75,225,88,65,146,189,218,67,164,26,121,65,98,103,198,218,67,88,14,160,65,17,216,218,67,64,10,204,65,70,214,218,67,60,180,242,65,98,222,212,218,67,26,4,6,66,146,189,218,67,48,221,25,
        66,146,189,218,67,188,73,46,66,98,145,189,218,67,27,47,59,66,0,96,218,67,133,107,90,66,39,81,219,67,78,98,95,66,99,109,88,153,156,67,242,82,134,66,98,29,10,156,67,240,103,133,66,123,68,155,67,176,242,133,66,190,223,154,67,88,121,135,66,98,141,103,154,
        67,68,75,137,66,186,137,154,67,127,42,140,66,79,125,154,67,199,75,144,66,98,243,77,154,67,232,251,159,66,220,116,154,67,193,10,179,66,253,100,154,67,199,139,194,66,98,46,93,154,67,162,197,201,66,194,165,154,67,72,97,211,66,34,27,154,67,199,203,216,66,
        98,237,156,153,67,238,188,221,66,28,122,152,67,109,231,224,66,44,34,151,67,209,98,226,66,98,73,44,147,67,0,192,230,66,146,72,144,67,236,81,222,66,11,34,144,67,35,91,208,66,98,62,21,144,67,51,179,203,66,191,58,144,67,215,163,196,66,191,58,144,67,248,83,
        190,66,98,191,58,144,67,211,141,176,66,191,58,144,67,78,98,165,66,191,58,144,67,207,247,151,66,98,191,58,144,67,191,223,147,66,248,126,144,67,144,2,137,66,74,167,143,67,189,180,134,66,98,41,55,143,67,19,131,133,66,162,144,142,67,222,36,134,66,80,104,
        142,67,111,82,134,66,98,217,89,142,67,209,98,134,66,154,68,142,67,220,57,134,66,41,55,142,67,111,82,134,66,98,250,41,141,67,113,61,136,66,141,114,141,67,197,32,146,66,141,114,141,67,132,128,153,66,98,141,114,141,67,66,96,166,66,141,114,141,67,159,47,
        185,66,141,114,141,67,226,58,197,66,98,141,114,141,67,222,164,205,66,188,63,141,67,89,185,211,66,180,163,141,67,248,83,218,66,98,157,202,141,67,215,227,220,66,33,59,142,67,232,187,224,66,194,128,142,67,203,97,226,66,98,26,250,142,67,0,64,229,66,229,123,
        144,67,148,216,233,66,178,72,145,67,58,52,235,66,98,201,113,147,67,54,222,238,66,235,44,151,67,15,173,238,66,225,37,153,67,136,150,235,66,98,145,56,154,67,116,232,233,66,63,176,155,67,32,69,229,66,172,55,156,67,126,255,225,66,98,221,47,157,67,126,255,
        219,66,84,94,157,67,245,189,210,66,84,94,157,67,52,115,201,66,98,84,94,157,67,208,119,188,66,84,94,157,67,240,39,167,66,84,94,157,67,234,166,154,66,98,244,93,157,67,104,209,146,66,190,159,157,67,131,0,136,66,88,153,156,67,242,82,134,66,99,109,23,153,
        142,67,100,59,66,66,98,84,19,143,67,217,206,77,66,91,116,144,67,225,250,89,66,193,170,145,67,184,158,94,66,98,232,187,147,67,37,134,102,66,201,22,151,67,102,102,101,66,17,136,153,67,133,235,96,66,98,64,197,154,67,227,165,94,66,154,185,155,67,74,140,90,
        66,107,188,155,67,72,225,83,66,98,15,189,155,67,64,53,82,66,160,122,155,67,52,179,76,66,170,65,155,67,164,112,75,66,98,182,163,154,67,164,240,71,66,229,128,153,67,246,40,75,66,180,72,152,67,219,249,76,66,98,164,192,150,67,131,64,79,66,125,255,148,67,
        219,249,79,66,157,223,147,67,219,249,76,66,98,192,218,146,67,137,65,74,66,225,250,145,67,117,19,69,66,186,121,145,67,47,221,60,66,98,131,0,145,67,209,34,53,66,237,252,144,67,94,186,38,66,26,255,144,67,59,223,27,66,98,83,3,145,67,110,18,6,66,34,251,144,
        67,20,174,223,65,26,255,144,67,93,186,176,65,98,71,1,145,67,34,219,152,65,32,0,145,67,171,28,129,65,104,97,145,67,200,161,99,65,98,67,43,146,67,120,20,36,65,225,250,148,67,144,24,4,65,161,181,151,67,118,233,22,65,98,249,174,152,67,200,161,29,65,134,102,
        154,67,28,133,43,65,151,174,154,67,212,163,10,65,98,51,243,154,67,184,159,214,64,110,50,154,67,47,94,182,64,20,62,153,67,2,172,160,64,98,8,236,151,67,233,167,130,64,192,218,149,67,204,206,111,64,176,114,148,67,18,4,130,64,98,115,104,147,67,206,120,137,
        64,199,251,145,67,30,6,165,64,147,72,145,67,84,100,191,64,98,77,98,144,67,119,63,225,64,114,24,143,67,30,176,34,65,94,202,142,67,30,176,56,65,98,49,136,142,67,196,75,75,65,213,72,142,67,186,116,117,65,76,55,142,67,64,96,138,65,98,139,12,142,67,90,143,
        175,65,76,55,142,67,246,83,215,65,76,55,142,67,40,220,4,66,98,43,55,142,67,233,38,27,66,121,9,142,67,160,154,52,66,23,153,142,67,100,59,66,66,99,109,246,136,223,67,168,6,188,66,98,193,90,222,67,174,7,184,66,183,115,220,67,84,35,181,66,185,238,218,67,
        236,209,178,66,98,242,50,217,67,140,44,176,66,144,114,215,67,44,135,174,66,122,217,213,67,228,37,171,66,98,106,1,213,67,42,92,169,66,124,212,211,67,169,134,165,66,32,165,211,67,140,44,161,66,98,175,103,211,67,50,136,155,66,184,227,211,67,189,52,152,66,
        7,81,212,67,206,12,150,66,98,87,174,213,67,85,35,143,66,103,214,216,67,175,71,142,66,236,129,219,67,200,75,144,66,98,213,184,220,67,202,54,145,66,119,142,221,67,94,143,146,66,23,25,222,67,22,174,144,66,98,93,95,222,67,90,185,143,66,23,137,222,67,178,
        242,140,66,101,123,222,67,219,78,139,66,98,111,66,222,67,211,162,132,66,150,19,218,67,6,22,132,66,62,186,216,67,110,103,132,66,98,105,113,215,67,65,181,132,66,209,210,214,67,100,16,133,66,78,242,213,67,34,240,133,66,98,6,145,212,67,93,79,135,66,92,15,
        212,67,126,255,135,66,94,42,211,67,140,236,138,66,98,250,110,210,67,111,82,141,66,51,243,209,67,228,165,143,66,180,136,209,67,148,152,146,66,98,47,93,209,67,211,205,147,66,94,42,209,67,105,209,148,66,243,13,209,67,74,12,150,66,98,175,66,208,67,96,229,
        158,66,145,221,208,67,15,45,167,66,34,235,209,67,21,174,172,66,98,177,29,210,67,64,181,173,66,147,104,210,67,254,20,175,66,91,175,210,67,203,33,176,66,98,224,154,211,67,35,155,179,66,165,155,212,67,39,241,180,66,120,217,213,67,187,9,183,66,98,46,61,214,
        67,40,177,183,66,118,158,214,67,54,94,184,66,97,0,215,67,58,244,184,66,98,81,72,216,67,134,235,186,66,7,108,217,67,42,220,188,66,62,165,218,67,64,181,190,66,98,240,114,219,67,9,236,191,66,173,71,220,67,44,7,193,66,109,242,220,67,15,237,194,66,98,248,
        238,221,67,107,188,197,66,97,224,222,67,93,143,202,66,2,246,222,67,208,247,207,66,98,78,13,223,67,159,239,213,66,158,106,222,67,87,206,217,66,90,207,221,67,67,160,220,66,98,2,230,220,67,174,220,224,66,248,142,219,67,178,242,226,66,47,200,217,67,176,135,
        227,66,98,55,132,214,67,31,154,228,66,233,145,212,67,22,174,223,66,246,115,211,67,242,103,216,66,98,101,246,210,67,197,53,213,66,166,182,210,67,224,164,209,66,142,210,209,67,56,30,209,66,98,93,154,207,67,220,206,207,66,72,60,208,67,177,71,217,66,7,172,
        208,67,15,2,221,66,98,44,162,209,67,253,62,229,66,183,222,211,67,87,163,235,66,165,59,214,67,134,128,237,66,98,126,154,216,67,193,95,239,66,165,139,219,67,255,169,238,66,95,133,221,67,210,247,235,66,98,242,253,221,67,120,83,235,66,220,84,222,67,163,26,
        234,66,154,196,222,67,230,229,232,66,98,138,172,223,67,224,100,230,66,238,183,224,67,11,44,226,66,60,42,225,67,195,138,222,66,98,253,84,225,67,160,47,221,66,177,109,225,67,158,4,219,66,138,140,225,67,136,43,217,66,98,224,250,225,67,32,154,210,66,40,220,
        225,67,171,134,201,66,17,67,225,67,13,215,196,66,98,145,221,224,67,250,190,193,66,139,28,224,67,213,248,189,66,246,136,223,67,168,6,188,66,99,109,0,64,244,67,74,12,118,65,98,66,64,244,67,68,139,88,65,150,99,244,67,238,124,35,65,82,88,244,67,70,182,13,
        65,98,25,68,244,67,136,22,205,64,74,252,243,67,54,94,146,64,125,207,242,67,128,106,148,64,98,16,56,241,67,3,43,151,64,92,95,241,67,50,8,12,65,92,95,241,67,195,245,68,65,98,92,95,241,67,144,194,159,65,92,95,241,67,52,51,209,65,92,95,241,67,197,160,5,66,
        98,92,95,241,67,72,225,13,66,137,145,241,67,86,142,43,66,194,21,241,67,132,192,44,66,98,212,168,240,67,206,204,45,66,180,200,239,67,148,152,32,66,198,91,239,67,167,27,27,66,98,16,200,238,67,222,164,19,66,92,79,238,67,116,232,12,66,241,210,237,67,247,
        40,7,66,98,85,142,235,67,242,167,216,65,110,162,233,67,234,251,166,65,108,151,231,67,62,180,102,65,98,28,42,231,67,7,43,81,65,147,136,230,67,27,217,54,65,36,246,229,67,4,0,26,65,98,108,215,229,67,235,251,19,65,124,207,229,67,143,108,11,65,73,172,229,
        67,22,131,4,65,98,167,86,229,67,206,75,231,64,57,180,228,67,40,176,174,64,40,60,228,67,16,172,160,64,98,159,186,227,67,112,145,145,64,65,16,227,67,161,153,153,64,46,253,226,67,100,143,154,64,98,85,30,225,67,211,161,177,64,1,235,225,67,176,28,46,65,69,
        214,225,67,153,67,127,65,98,16,200,225,67,196,245,155,65,69,214,225,67,202,118,192,65,69,214,225,67,227,122,210,65,98,69,214,225,67,157,196,225,65,231,219,225,67,210,34,249,65,69,214,225,67,193,202,1,66,98,182,195,225,67,103,230,19,66,108,199,225,67,
        236,81,52,66,210,189,225,67,210,34,73,66,98,245,184,225,67,75,12,84,66,46,237,225,67,114,61,95,66,83,179,226,67,134,235,96,66,98,67,155,228,67,75,12,101,66,237,172,228,67,38,134,85,66,200,182,228,67,64,53,76,66,98,186,217,228,67,187,201,42,66,200,182,
        228,67,234,38,13,66,200,182,228,67,89,57,221,65,98,200,182,228,67,163,69,205,65,198,107,228,67,116,104,150,65,239,231,228,67,7,129,144,65,98,196,48,229,67,50,8,141,65,184,126,229,67,224,79,149,65,100,123,229,67,173,28,149,65,98,231,171,229,67,234,38,
        152,65,82,184,229,67,9,172,152,65,37,246,229,67,236,81,158,65,98,31,37,231,67,160,26,186,65,168,86,232,67,113,61,216,65,6,81,233,67,58,180,242,65,98,192,186,234,67,189,116,12,66,147,24,236,67,246,40,28,66,88,137,237,67,187,73,46,66,98,170,97,238,67,146,
        237,56,66,86,62,239,67,138,65,67,66,141,7,240,67,113,189,77,66,98,96,133,240,67,205,76,84,66,84,19,241,67,50,136,94,66,34,11,242,67,134,235,96,66,98,153,153,242,67,200,75,98,66,159,106,243,67,208,119,98,66,145,221,243,67,79,98,95,66,98,196,208,244,67,
        42,220,88,66,20,62,244,67,204,161,53,66,0,64,244,67,77,55,43,66,98,221,68,244,67,63,181,13,66,20,62,244,67,137,65,183,65,0,64,244,67,74,12,118,65,99,109,223,231,3,68,133,107,52,66,98,223,231,3,68,182,243,32,66,100,11,4,68,35,91,10,66,194,133,3,68,90,
        100,6,66,98,221,60,3,68,94,58,4,66,237,172,2,68,121,105,5,66,44,58,2,68,190,159,5,66,98,118,214,1,68,217,206,5,66,124,95,1,68,40,92,5,66,114,56,1,68,243,253,10,66,98,71,9,1,68,229,208,17,66,167,94,1,68,83,227,22,66,102,142,1,68,204,204,24,66,98,120,225,
        1,68,196,32,28,66,245,64,2,68,75,183,26,66,100,107,2,68,7,44,30,66,98,139,164,2,68,234,209,34,66,30,141,2,68,249,126,42,66,141,119,2,68,28,90,49,66,98,123,92,2,68,243,253,57,66,145,29,2,68,231,123,65,66,166,203,1,68,198,75,69,66,98,162,133,1,68,85,142,
        72,66,113,37,1,68,92,143,75,66,70,214,0,68,206,247,76,66,98,31,21,255,67,16,216,82,66,124,84,252,67,93,186,71,66,122,137,251,67,79,141,58,66,98,192,159,250,67,102,102,43,66,185,14,251,67,83,227,19,66,185,14,251,67,31,176,242,65,98,185,14,251,67,60,10,
        206,65,255,228,250,67,18,174,166,65,185,14,251,67,70,225,139,65,98,175,55,251,67,107,18,99,65,66,16,252,67,144,24,62,65,154,249,252,67,50,94,44,65,98,43,231,253,67,1,86,26,65,60,79,255,67,150,153,13,65,230,152,0,68,214,206,19,65,98,150,219,0,68,52,137,
        21,65,179,117,1,68,44,221,34,65,87,142,1,68,230,38,35,65,98,50,0,2,68,15,131,36,65,17,80,2,68,11,45,28,65,44,95,2,68,52,137,7,65,98,11,119,2,68,157,112,205,64,116,248,1,68,51,180,176,64,30,130,1,68,110,147,160,64,98,130,109,1,68,137,194,157,64,122,89,
        1,68,6,2,151,64,206,68,1,68,229,81,148,64,98,212,205,0,68,51,180,132,64,41,241,255,67,30,135,126,64,56,46,255,67,127,235,129,64,98,89,30,255,67,166,28,130,64,83,13,255,67,117,20,130,64,17,253,254,67,127,235,129,64,98,228,122,253,67,39,51,123,64,111,247,
        251,67,150,196,164,64,13,39,251,67,74,141,203,64,98,175,236,250,67,146,110,214,64,33,197,250,67,21,47,229,64,251,147,250,67,23,90,240,64,98,114,34,250,67,182,30,5,65,255,169,249,67,157,26,17,65,116,109,249,67,10,2,29,65,98,196,138,248,67,4,129,73,65,
        24,46,248,67,249,126,134,65,24,46,248,67,48,8,169,65,98,24,46,248,67,247,83,194,65,8,38,248,67,38,49,227,65,24,46,248,67,67,139,2,66,98,8,54,248,67,90,100,18,66,212,34,248,67,81,56,35,66,106,70,248,67,116,147,48,66,98,24,110,248,67,192,74,63,66,208,92,
        249,67,147,152,77,66,212,98,250,67,52,222,83,66,98,11,236,250,67,232,38,87,66,208,92,251,67,1,171,90,66,169,235,251,67,109,18,93,66,98,183,24,253,67,220,36,98,66,106,230,254,67,70,225,100,66,45,79,0,68,117,190,100,66,98,180,229,0,68,232,166,100,66,108,
        68,1,68,58,95,98,66,70,179,1,68,213,35,96,66,98,195,234,1,68,29,5,95,66,64,74,2,68,160,69,92,66,127,119,2,68,215,78,92,66,98,80,218,2,68,63,96,92,66,211,10,3,68,160,69,99,66,180,133,3,68,213,35,96,66,98,6,25,4,68,96,101,92,66,223,231,3,68,248,211,66,
        66,223,231,3,68,133,107,52,66,99,109,20,206,158,67,254,212,70,66,98,90,148,157,67,37,6,52,66,219,9,158,67,230,80,12,66,219,9,158,67,76,55,221,65,98,219,9,158,67,23,217,163,65,248,227,157,67,91,100,115,65,194,181,158,67,95,186,59,65,98,233,38,159,67,204,
        161,29,65,200,198,159,67,145,194,253,64,131,160,160,67,145,194,209,64,98,47,141,161,67,233,251,161,64,235,241,162,67,239,81,120,64,174,167,164,67,163,26,95,64,98,242,146,166,67,233,208,66,64,16,248,168,67,77,12,98,64,72,129,170,67,98,229,172,64,98,125,
        127,171,67,229,165,211,64,113,205,172,67,177,114,24,65,56,73,173,67,181,200,62,65,98,72,33,174,67,35,219,128,65,140,220,173,67,111,18,201,65,140,220,173,67,186,73,247,65,98,140,220,173,67,194,117,17,66,34,32,174,67,61,10,45,66,62,122,173,67,90,100,62,
        66,98,181,248,172,67,145,237,75,66,250,158,171,67,102,102,87,66,250,30,170,67,16,216,93,66,98,225,186,169,67,30,133,95,66,158,223,168,67,198,75,97,66,25,52,168,67,176,114,98,66,98,231,43,164,67,72,97,105,66,63,69,160,67,223,79,93,66,20,206,158,67,254,
        212,70,66,99,109,246,72,166,67,6,129,78,66,98,123,180,167,67,225,250,77,66,117,179,169,67,2,171,71,66,33,80,170,67,39,177,64,66,98,176,146,170,67,88,185,61,66,185,254,170,67,117,147,51,66,90,20,171,67,25,132,45,66,98,88,73,171,67,236,209,30,66,90,20,
        171,67,27,175,11,66,90,20,171,67,46,178,242,65,98,90,20,171,67,58,180,206,65,137,81,171,67,190,116,172,65,90,20,171,67,242,210,136,65,98,233,198,170,67,206,204,54,65,94,74,168,67,194,202,253,64,192,58,165,67,46,178,13,65,98,167,166,163,67,17,88,21,65,
        196,48,162,67,222,36,40,65,112,125,161,67,222,36,78,65,98,225,250,160,67,218,206,105,65,221,212,160,67,2,43,144,65,137,209,160,67,121,233,162,65,98,127,202,160,67,176,114,205,65,137,209,160,67,176,114,0,66,137,209,160,67,104,145,25,66,98,137,209,160,
        67,45,50,38,66,184,222,160,67,77,226,51,66,74,76,161,67,241,82,59,66,98,137,97,161,67,149,195,60,66,41,172,161,67,192,202,62,66,191,223,161,67,38,177,64,66,98,150,19,162,67,177,157,66,66,160,58,162,67,20,174,68,66,209,114,162,67,97,16,70,66,98,82,104,
        163,67,111,18,76,66,121,169,164,67,160,26,79,66,246,72,166,67,6,129,78,66,99,109,35,203,184,67,88,121,135,66,98,115,56,184,67,74,204,132,66,221,164,182,67,47,157,133,66,45,210,181,67,164,240,133,66,98,25,4,181,67,144,66,134,66,98,0,180,67,25,68,134,66,
        49,24,180,67,92,79,139,66,98,123,52,180,67,104,81,145,66,170,225,181,67,229,144,142,66,205,76,182,67,76,55,146,66,98,178,125,182,67,60,223,147,66,205,76,182,67,248,211,152,66,205,76,182,67,189,244,156,66,98,205,76,182,67,44,135,167,66,242,130,182,67,
        189,180,181,66,64,101,182,67,48,221,191,66,98,183,83,182,67,236,209,197,66,64,101,182,67,214,56,206,66,64,101,182,67,251,126,214,66,98,64,101,182,67,20,3,221,66,191,255,181,67,153,174,233,66,148,248,182,67,220,249,235,66,98,189,116,183,67,54,30,237,66,
        228,101,184,67,23,153,236,66,177,178,184,67,142,151,235,66,98,36,43,185,67,138,1,234,66,122,41,185,67,201,54,227,66,81,45,185,67,128,42,222,66,98,28,63,185,67,73,33,199,66,28,63,185,67,140,108,178,66,157,20,185,67,245,125,158,66,98,76,7,185,67,133,43,
        152,66,43,151,185,67,176,50,139,66,35,203,184,67,88,121,135,66,99,109,166,203,212,67,254,212,70,66,98,160,202,211,67,4,86,86,66,51,195,209,67,176,242,95,66,45,242,206,67,20,174,97,66,98,45,162,205,67,243,125,98,66,84,19,204,67,20,174,97,66,59,63,202,
        67,20,174,97,66,98,196,160,200,67,26,175,97,66,20,174,198,67,90,100,99,66,157,31,198,67,22,217,93,66,98,30,197,197,67,241,82,90,66,235,193,197,67,210,77,77,66,79,189,197,67,204,76,69,66,98,140,183,197,67,114,104,59,66,167,214,197,67,54,9,52,66,3,214,
        197,67,231,251,43,66,98,202,209,197,67,214,35,2,66,206,215,197,67,71,225,171,65,79,189,197,67,220,79,41,65,98,233,182,197,67,110,147,252,64,99,203,197,67,176,243,185,64,163,80,198,67,209,163,160,64,98,159,26,199,67,213,122,116,64,147,152,201,67,213,249,
        142,64,186,185,202,67,107,61,142,64,98,141,183,202,67,107,61,142,64,229,192,202,67,211,206,135,64,110,210,202,67,142,24,136,64,98,122,132,205,67,222,165,147,64,108,167,207,67,177,243,145,64,239,87,209,67,208,120,197,64,98,241,82,210,67,134,108,227,64,
        32,80,211,67,40,135,16,65,233,6,212,67,100,102,44,65,98,204,172,212,67,200,161,69,65,163,128,213,67,238,167,129,65,147,168,213,67,112,61,155,65,98,55,201,213,67,98,16,176,65,100,171,213,67,110,18,196,65,147,168,213,67,178,157,216,65,98,66,160,213,67,
        129,149,8,66,43,7,214,67,193,202,51,66,166,203,212,67,254,212,70,66,99,109,131,224,210,67,147,152,15,66,98,131,224,210,67,98,16,245,65,235,225,210,67,1,43,208,65,131,224,210,67,180,200,179,65,98,92,223,210,67,24,4,152,65,14,77,210,67,65,96,109,65,137,
        161,209,67,161,69,84,65,98,76,87,208,67,108,231,35,65,106,108,205,67,230,251,19,65,84,147,201,67,61,10,29,65,98,231,123,201,67,124,63,29,65,176,130,201,67,192,202,31,65,225,122,201,67,147,24,32,65,98,139,108,201,67,8,172,32,65,174,87,201,67,16,88,31,
        65,186,73,201,67,147,24,32,65,98,12,178,200,67,18,131,40,65,10,167,200,67,190,159,62,65,211,157,200,67,196,32,78,65,98,39,145,200,67,251,169,99,65,70,182,200,67,167,198,123,65,70,182,200,67,118,190,133,65,98,115,184,200,67,136,65,180,65,70,182,200,67,
        198,75,221,65,70,182,200,67,132,235,7,66,98,70,182,200,67,214,35,32,66,166,171,200,67,44,50,41,66,70,182,200,67,34,219,60,66,98,29,186,200,67,28,218,67,66,213,168,200,67,155,68,73,66,33,0,201,67,151,110,75,66,98,185,110,201,67,26,47,78,66,232,187,203,
        67,118,62,77,66,113,189,204,67,206,247,76,66,98,172,12,206,67,159,154,76,66,68,187,207,67,229,208,74,66,220,73,208,67,202,33,73,66,98,243,242,208,67,202,33,71,66,118,211,209,67,50,179,65,66,107,28,210,67,89,100,62,66,98,57,244,210,67,141,151,52,66,131,
        224,210,67,170,113,35,66,131,224,210,67,147,152,15,66,99,109,82,88,116,67,57,52,76,66,98,232,219,116,67,0,0,83,66,88,153,117,67,86,142,90,66,160,186,116,67,178,157,94,66,98,156,68,116,67,155,196,96,66,144,34,115,67,127,234,97,66,94,218,113,67,127,234,
        96,66,98,239,71,112,67,45,178,95,66,94,186,111,67,137,193,88,66,2,43,111,67,10,87,82,66,98,14,45,110,67,24,4,71,66,12,226,108,67,129,149,57,66,115,232,107,67,18,131,45,66,98,209,98,107,67,86,14,39,66,211,13,107,67,18,3,31,66,238,252,105,67,208,162,28,
        66,98,54,254,104,67,120,105,26,66,158,79,101,67,52,222,27,66,52,179,98,67,52,222,27,66,98,167,91,97,67,52,222,27,66,152,46,92,67,17,3,26,66,148,56,91,67,107,103,29,66,98,179,29,90,67,203,76,33,66,222,164,90,67,95,229,56,66,222,164,90,67,193,117,65,66,
        98,222,164,90,67,240,210,78,66,171,145,91,67,29,133,98,66,183,147,87,67,125,234,96,66,98,191,31,84,67,41,135,95,66,255,20,85,67,219,36,72,66,255,20,85,67,186,244,53,66,98,255,20,85,67,67,54,8,66,255,20,85,67,147,110,183,65,255,20,85,67,84,186,59,65,98,
        255,20,85,67,86,229,18,65,226,250,84,67,227,83,207,64,50,168,85,67,90,18,179,64,98,134,107,86,67,119,108,147,64,236,241,87,67,176,32,152,64,134,235,88,67,106,106,148,64,98,13,2,91,67,155,114,140,64,77,215,93,67,38,223,147,64,116,200,96,67,106,106,148,
        64,98,132,0,101,67,212,38,149,64,216,35,105,67,20,92,139,64,253,73,108,67,208,208,166,64,98,214,216,110,67,114,22,189,64,97,5,112,67,212,38,213,64,247,168,113,67,118,149,7,65,98,161,186,115,67,28,49,44,65,186,94,117,67,36,221,90,65,34,176,117,67,0,129,
        144,65,98,138,1,118,67,12,131,179,65,106,209,117,67,184,159,226,65,175,39,116,67,213,249,254,65,98,38,198,115,67,110,189,2,66,193,234,114,67,46,136,7,66,42,60,114,67,247,254,10,66,98,194,138,113,67,40,135,14,66,13,130,112,67,63,224,16,66,105,81,112,67,
        110,189,21,66,98,17,24,112,67,4,129,27,66,44,7,113,67,255,170,35,66,82,120,113,67,124,234,40,66,98,2,139,114,67,76,183,53,66,176,50,115,67,49,8,61,66,82,88,116,67,57,52,76,66,99,109,178,221,108,67,8,172,56,65,98,213,152,106,67,2,43,31,65,6,225,102,67,
        26,47,35,65,242,178,98,67,26,47,35,65,98,168,70,96,67,26,47,35,65,156,68,93,67,221,36,32,65,146,45,92,67,137,65,38,65,98,105,17,92,67,47,221,38,65,37,230,91,67,188,116,37,65,68,203,91,67,137,65,38,65,98,76,87,90,67,198,75,49,65,91,164,90,67,88,57,84,
        65,91,164,90,67,74,12,118,65,98,91,164,90,67,203,161,138,65,91,164,90,67,232,251,149,65,91,164,90,67,129,149,170,65,98,91,164,90,67,225,122,194,65,187,233,89,67,84,227,251,65,17,56,91,67,242,82,3,66,98,115,40,92,67,52,51,7,66,191,255,96,67,99,144,5,66,
        177,178,98,67,197,160,5,66,98,195,149,101,67,107,188,5,66,134,235,104,67,207,119,6,66,85,35,107,67,242,82,3,66,98,3,203,108,67,226,250,0,66,234,38,110,67,38,6,245,65,50,200,110,67,38,6,235,65,98,55,73,113,67,29,90,195,65,20,110,112,67,131,192,96,65,178,
        221,108,67,8,172,56,65,99,109,242,34,193,67,100,59,66,66,98,144,98,193,67,155,68,72,66,205,44,194,67,225,122,87,66,185,206,193,67,129,21,93,66,98,253,105,193,67,135,22,99,66,179,253,191,67,131,192,98,66,132,80,191,67,80,141,91,66,98,97,229,190,67,148,
        24,87,66,255,164,190,67,199,75,77,66,36,91,190,67,4,214,70,66,98,126,255,189,67,35,219,62,66,165,176,189,67,209,162,56,66,97,101,189,67,35,91,49,66,98,58,36,189,67,80,13,43,66,113,221,188,67,195,245,31,66,142,87,188,67,115,104,29,66,98,140,140,187,67,
        68,139,25,66,68,235,185,67,88,185,27,66,210,178,184,67,60,223,27,66,98,190,20,184,67,177,242,27,66,251,78,181,67,11,215,25,66,206,220,180,67,115,104,29,66,98,40,129,180,67,138,65,32,66,112,146,180,67,31,5,53,66,210,146,180,67,92,143,58,66,98,255,148,
        180,67,235,81,76,66,196,69,181,67,219,121,99,66,138,241,178,67,133,235,96,66,98,44,183,178,67,8,172,96,66,228,85,178,67,180,200,95,66,81,45,178,67,184,158,94,66,98,220,105,177,67,61,10,89,66,3,203,177,67,186,73,60,66,3,203,177,67,186,73,46,66,98,3,203,
        177,67,22,89,2,66,3,203,177,67,182,243,179,65,3,203,177,67,152,153,53,65,98,3,203,177,67,138,108,21,65,44,183,177,67,150,153,209,64,124,20,178,67,107,18,179,64,98,220,137,178,67,142,237,140,64,65,85,179,67,76,141,151,64,18,24,180,67,123,106,148,64,98,
        108,12,182,67,74,98,140,64,204,145,185,67,0,86,138,64,26,196,187,67,123,106,148,64,98,65,245,189,67,148,110,158,64,81,77,191,67,25,90,208,64,12,119,192,67,33,219,19,65,98,190,132,193,67,162,112,59,65,184,19,194,67,236,124,121,65,8,49,194,67,89,100,161,
        65,98,76,76,194,67,81,184,195,65,178,2,194,67,26,47,233,65,244,34,193,67,198,203,1,66,98,0,245,192,67,255,127,4,66,224,196,192,67,245,168,7,66,13,119,192,67,100,59,10,66,98,156,41,192,67,198,203,12,66,136,123,191,67,34,91,17,66,216,104,191,67,12,130,
        22,66,98,204,86,191,67,24,132,27,66,69,192,191,67,67,139,35,66,77,252,191,67,133,235,40,66,98,137,97,192,67,201,246,49,66,0,192,192,67,47,221,56,66,242,34,193,67,100,59,66,66,99,109,156,164,190,67,152,110,90,65,98,68,139,189,67,84,227,31,65,164,176,187,
        67,27,47,35,65,176,178,184,67,27,47,35,65,98,246,120,183,67,27,47,35,65,125,175,181,67,60,223,31,65,151,62,181,67,138,65,38,65,98,30,37,181,67,46,178,39,65,26,31,181,67,218,206,39,65,210,13,181,67,224,79,41,65,98,52,126,180,67,60,223,53,65,130,176,180,
        67,128,106,106,65,99,171,180,67,187,73,135,65,98,81,168,180,67,244,253,146,65,222,143,180,67,85,227,166,65,175,146,180,67,134,235,185,65,98,128,149,180,67,204,161,203,65,212,136,180,67,242,82,0,66,171,220,180,67,236,81,3,66,98,251,73,181,67,119,62,7,
        66,13,77,183,67,64,181,5,66,179,104,184,67,191,159,5,66,98,52,206,186,67,164,112,5,66,89,212,188,67,111,18,6,66,179,248,189,67,110,231,251,65,98,193,170,191,67,205,204,227,65,176,226,191,67,150,67,142,65,156,164,190,67,152,110,90,65,99,109,29,154,191,
        67,168,198,137,66,98,2,187,192,67,8,44,134,66,193,106,194,67,219,185,131,66,168,150,196,67,6,65,131,66,98,248,211,198,67,156,196,130,66,35,75,201,67,119,126,133,66,180,136,202,67,192,138,138,66,98,78,18,203,67,94,186,140,66,125,207,203,67,133,235,143,
        66,94,42,204,67,225,250,146,66,98,170,97,205,67,207,119,157,66,151,238,204,67,231,59,176,66,151,238,204,67,41,220,191,66,98,151,238,204,67,170,49,204,66,18,3,205,67,57,116,216,66,176,66,204,67,20,238,222,66,98,10,167,203,67,126,42,228,66,33,64,202,67,
        170,49,233,66,26,207,200,67,135,150,235,66,98,143,18,199,67,207,119,238,66,102,150,196,67,221,228,238,66,192,122,194,67,190,31,237,66,98,38,65,192,67,0,64,235,66,32,112,190,67,116,83,229,66,167,150,189,67,96,101,221,66,98,204,188,188,67,176,114,213,66,
        149,3,189,67,6,129,194,66,149,3,189,67,30,69,182,66,98,149,3,189,67,104,145,169,66,200,198,188,67,98,144,154,66,206,199,189,67,200,54,146,66,98,132,235,189,67,210,13,145,66,97,64,190,67,118,126,143,66,106,140,190,67,118,254,141,66,109,168,38,202,67,92,
        143,189,66,98,168,38,202,67,135,214,178,66,148,104,202,67,211,141,169,66,168,38,202,67,90,164,159,66,98,109,231,201,67,141,23,150,66,59,143,200,67,203,33,144,66,236,129,198,67,72,97,142,66,98,195,197,197,67,6,193,141,66,58,100,196,67,156,4,142,66,111,
        210,195,67,72,97,142,66,98,179,157,194,67,228,37,143,66,226,138,193,67,140,44,145,66,203,241,192,67,131,192,147,66,98,8,124,191,67,68,11,154,66,183,227,191,67,110,18,168,66,183,227,191,67,139,108,183,66,98,183,227,191,67,10,87,189,66,183,227,191,67,254,
        20,195,66,183,227,191,67,133,235,199,66,98,183,227,191,67,188,52,206,66,54,190,191,67,172,28,212,66,222,20,192,67,174,7,216,66,98,251,62,192,67,145,237,217,66,130,165,192,67,70,182,219,66,161,10,193,67,102,102,221,66,98,81,253,193,67,57,116,225,66,165,
        112,195,67,61,202,227,66,102,91,197,67,186,137,227,66,98,133,128,198,67,84,99,227,66,247,216,199,67,174,199,225,66,131,133,200,67,182,179,223,66,98,221,116,202,67,219,185,217,66,168,38,202,67,190,223,203,66,168,38,202,67,92,143,189,66,99,109,240,55,177,
        67,88,185,185,66,98,240,55,177,67,197,224,198,66,113,141,177,67,184,30,215,66,226,90,176,67,129,21,224,66,98,177,226,175,67,147,152,227,66,238,28,174,67,127,170,233,66,126,255,172,67,63,53,235,66,98,241,71,171,67,247,147,237,66,110,103,169,67,65,32,237,
        66,5,38,167,67,65,32,237,66,98,140,140,165,67,190,31,237,66,144,66,162,67,153,89,238,66,228,149,161,67,241,210,234,66,98,62,10,161,67,200,246,231,66,107,76,161,67,85,206,215,66,107,76,161,67,114,168,210,66,98,107,76,161,67,16,216,195,66,107,76,161,67,
        69,54,175,66,107,76,161,67,36,6,160,66,98,107,76,161,67,69,54,156,66,197,240,160,67,206,119,139,66,228,149,161,67,165,219,135,66,98,64,213,161,67,224,122,134,66,152,62,162,67,241,82,134,66,140,188,162,67,38,241,133,66,98,146,77,164,67,212,184,132,66,
        142,71,167,67,120,41,133,66,197,16,169,67,216,142,133,66,98,215,163,171,67,196,32,134,66,236,97,173,67,245,168,136,66,56,185,174,67,171,156,141,66,98,165,80,176,67,237,124,147,66,54,62,177,67,163,48,157,66,66,80,177,67,99,59,169,66,98,76,87,177,67,2,
        235,173,66,240,55,177,67,213,184,179,66,240,55,177,67,88,185,185,66,99,109,100,171,173,67,131,128,153,66,98,229,144,172,67,203,225,146,66,49,232,170,67,29,90,144,66,139,76,168,67,252,233,143,66,98,233,118,167,67,162,197,143,66,24,116,165,67,174,135,143,
        66,71,241,164,67,199,75,144,66,98,208,226,164,67,203,97,144,66,216,206,164,67,207,55,144,66,32,192,164,67,199,75,144,66,98,102,70,164,67,164,240,144,66,149,51,164,67,86,78,148,66,14,45,164,67,205,12,150,66,98,219,41,164,67,223,207,150,66,229,48,164,67,
        43,135,155,66,14,45,164,67,135,86,157,66,98,106,28,164,67,205,204,164,66,14,45,164,67,18,3,167,66,14,45,164,67,229,16,173,66,98,14,45,164,67,204,76,181,66,245,72,164,67,63,181,197,66,96,69,164,67,106,252,202,66,98,57,68,164,67,108,231,204,66,123,36,164,
        67,24,132,208,66,14,45,164,67,118,126,214,66,98,77,50,164,67,210,77,218,66,55,41,164,67,59,95,224,66,135,118,164,67,52,158,225,66,98,151,254,164,67,85,206,227,66,0,32,167,67,144,237,226,66,221,100,168,67,154,196,226,66,98,20,46,171,67,0,107,226,66,219,
        57,173,67,187,52,224,66,80,13,174,67,66,203,216,66,98,156,116,174,67,13,45,213,66,158,111,174,67,93,250,207,66,158,111,174,67,105,252,202,66,98,158,111,174,67,134,150,197,66,158,111,174,67,121,20,192,66,158,111,174,67,11,66,187,66,98,158,111,174,67,244,
        189,175,66,90,212,174,67,207,119,160,66,100,171,173,67,131,128,153,66,99,101,0,0 };
    
    Path path;
    path.loadPathFromData (pathData, sizeof (pathData));

    return path;
}



//==============================================================================
//==============================================================================
SpectrumViewer::FrequencyCaption::FrequencyCaption()
: sampleRate {44100.0}
{
    for (int i = 0; i < numberOfFrequenciesToPlot; ++i)
    {
        // Generate the frequencyString...
        const int frequency = frequenciesToPlot[i];
        String frequencyString;
        if (frequency % 1000 == 0)
        {
            frequencyString = String (frequency/1000) + "K";
        }
        else
        {
            frequencyString = String (frequency);
        }
        
        // ... and put it into a new Label.
        Label* frequencyLabel = new Label();
        frequencyLabel->setText(frequencyString, dontSendNotification);
        
        // Figuring out the width and height of the text and setting the Label accordingly.
        Font labelFont = Font(12.0f);
        frequencyLabel->setFont(labelFont);
        frequencyLabel->setColour (Label::textColourId, Colours::lightgoldenrodyellow);
        const int textWidth = labelFont.getStringWidth (frequencyLabel->getText());
        const int textHeight = labelFont.getHeight();
        frequencyLabel->setBorderSize (BorderSize<int>(0,1,0,1));
        frequencyLabel->setSize(textWidth + 2, textHeight);
        
        frequencyLabels.add(frequencyLabel);
        addAndMakeVisible(frequencyLabel);
    }
}

SpectrumViewer::FrequencyCaption::~FrequencyCaption()
{
}

void SpectrumViewer::FrequencyCaption::paint (Graphics& g)
{
    // Background
    // ----------
    g.fillAll (Colours::black); // Clear the background
    
    // Labels and lines
    // ----------------
    static float lineLength = 5.0f;
    g.setColour (Colours::lightgoldenrodyellow);
    
    for (int i = numberOfFrequenciesToPlot - 1; i >= 0; --i)
    {
        // Figure out the line position...
        const double proportion = frequenciesToPlot[i] / (sampleRate * 0.5);
        int xPosOfLine = logTransformInRange0to1 (proportion) * getWidth();
        
        // ...and the label position.
        const int widthOfLabel = frequencyLabels[i]->getWidth();
        frequencyLabels[i]->setTopLeftPosition(xPosOfLine - widthOfLabel / 2, lineLength);
        
        // Do labels overlap? Hide if neccessary
        frequencyLabels[i]->setVisible (true);
        for (int j = i+1; j < numberOfFrequenciesToPlot; ++j)
        {
            if (frequencyLabels[j]->isVisible()
                && frequencyLabels[j]->getX() <= frequencyLabels[i]->getX() + frequencyLabels[i]->getWidth())
            {
                frequencyLabels[i]->setVisible(false);
                break;
            }
        }
        // Only show the 20K label if it isn't truncated by the border of the whole component.
        if (i == numberOfFrequenciesToPlot - 1)
        {
            if (frequencyLabels[i]->getX() + frequencyLabels[i]->getWidth() > getWidth())
            {
                frequencyLabels[i]->setVisible(false);
            }
        }
        
        if (frequencyLabels[i]->isVisible())
        {
            g.drawVerticalLine(xPosOfLine, 0.0f, lineLength);
        }
    }
}

void SpectrumViewer::FrequencyCaption::resized()
{
}

void SpectrumViewer::FrequencyCaption::setSampleRate (double newSampleRate)
{
    sampleRate = newSampleRate;
}

#endif // JUCE_MAC || JUCE_IOS || DROWAUDIO_USE_FFTREAL
